@@ -20,6 +20,22 @@
    If not, write to the Free Software Foundation,
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
+/*
+    This file, coming from the source code of GNU DIFF,
+    has been modified by Ivano Primi  <ivprimi@libero.it>
+    so that it could be merged into the source code of Numdiff.
+
+    Numdiff is free software; you can redistribute it and/or modify
+    it under the terms of the GNU General Public License as published by
+    the Free Software Foundation; either version 3 of the License, or
+    (at your option) any later version.
+
+    Numdiff is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+*/
+
 #include "numdiff.h"
 #include <cmpbuf.h>
 #include <setmode.h>
@@ -229,6 +245,7 @@ find_and_hash_each_line (struct file_data *current, char* ifs, int index, argsli
   lin eqs_alloc = equivs_alloc;
   char const *suffix_begin = current->suffix_begin;
   char const *bufend = FILE_BUFFER (current) + current->buffered;
+  const unsigned long fieldno_upper_limit = FIELDMASK_SIZE*8;
 
   while ((char const *) p < suffix_begin)
     {
@@ -240,7 +257,7 @@ find_and_hash_each_line (struct file_data *current, char* ifs, int index, argsli
       /* First we have to initialize `p' so that it points */
       /* to the first character of the first field */
       for (h = 0; *p != '\n' && (strchr (ifs, *p)); p++);
-      for (fieldno = 0; *p != '\n'; fieldno++)
+      for (fieldno = 0; fieldno < fieldno_upper_limit && *p != '\n'; fieldno++)
         {
           /* 
 	     The first time we enter in this cycle we have
@@ -330,6 +347,13 @@ find_and_hash_each_line (struct file_data *current, char* ifs, int index, argsli
           for (*eof = c; *p != '\n' && (strchr (ifs, *p)); p++);
           /* And now to the begin of the next field */
         }
+      if (fieldno >= fieldno_upper_limit)
+	{
+	  fprintf (stderr, _("***  Fatal error occurred in function %s:\n%s"),
+		   __FUNCTION__,
+		   _("***  a very long line has been encountered which contains\n***  too many fields to be correctly handled\n"));
+	  exit (EXIT_TROUBLE);
+	}
       p++;
 
       /*    hashing_done:; */
@@ -337,23 +361,10 @@ find_and_hash_each_line (struct file_data *current, char* ifs, int index, argsli
       bucket = &buckets[h % nbuckets];
       length = (char const *) p - ip - 1;
 
-#ifdef _DEBUG_
+#ifdef _DEBUG_HASHING_
       fprintf (stderr, "File: \"%s\"\nLine: \"%.*s\"\nHash: %u\n", 
 	       current->name, length, ip, h);
-#endif /* _DEBUG_ */
-
-      if ((char const *) p == bufend
-	  && current->missing_newline)
-	{
-	  /* This line is incomplete.  If this is significant,
-	     put the line into buckets[-1].  */
-	  if (ignore_white_space < IGNORE_SPACE_CHANGE)
-	    bucket = &buckets[-1];
-
-	  /* Omit the inserted newline when computing linbuf later.  */
-	  p--;
-	  bufend = suffix_begin = (char const *) p;
-	}
+#endif /* _DEBUG_HASHING_ */
 
       for (i = *bucket;  ;  i = eqs[i].next)
 	if (!i)
@@ -437,7 +448,12 @@ find_and_hash_each_line (struct file_data *current, char* ifs, int index, argsli
       linbuf[line] = (char const *) p;
 
       if ((char const *) p == bufend)
-	break;
+	{
+	  /* If the last line is incomplete, don't count its appended newline.  */
+	  if ((current->missing_newline))
+	    linbuf[line]--;
+	  break;
+	}
 
       if ((suppress_common_lines))
 	break;
@@ -456,6 +472,33 @@ find_and_hash_each_line (struct file_data *current, char* ifs, int index, argsli
   equivs = eqs;
   equivs_alloc = eqs_alloc;
   equivs_index = eqs_index;
+
+#ifdef _DEBUG_HASHING_
+  if (1)
+    {
+      size_t n;
+      lin j;
+
+      fprintf (stderr, "buckets[-1] = %ld\n", buckets[-1]);
+      for (n = 0; n < nbuckets; n++)
+	fprintf (stderr, "buckets[%zu] = %ld\n", n, buckets[n]);
+      for (n = 0; n < equivs_alloc; n++)
+	fprintf (stderr, "equivs[%zu]  = {next = %d, hash = %lu, line = %s, length = %zu, file = %d}\n", n,
+		 equivs[n].next, equivs[n].hash, equivs[n].line, equivs[n].length, equivs[n].file);
+
+      char const **linbuf = current->linbuf;
+      lin alloc_lines = current->alloc_lines;
+      fprintf (stderr, "current->linbuf_base = %ld, current->buffered_lines = %ld,\n current->valid_lines = %ld, current->alloc_lines = %ld\n",
+	       current->linbuf_base, current->buffered_lines, current->valid_lines, current->alloc_lines);
+      for ( j = current->linbuf_base; j < current->alloc_lines; j++ )
+	fprintf (stderr, "current->linbuf[%ld] = %s\n", j, current->linbuf[j]);
+      fprintf (stderr, "  current->equivs = ");
+      for ( j = 0; j < current->buffered_lines; j++ )
+	fprintf (stderr, "%ld ", current->equivs[j]);
+      fputs ("\n", stderr);
+    }
+#endif /* _DEBUG_HASHING_ */
+
 }
 
 /* Prepare the text.  Make sure the text end is initialized.
@@ -594,10 +637,11 @@ find_identical_ends (struct file_data filevec[])
       beg0 = filevec[0].prefix_end + (n0 < n1 ? 0 : n0 - n1);
 
       /* Scan back until chars don't match or we reach that point.  */
-      for (; p0 != beg0; p0--, p1--)
-	if (*p0 != *p1)
+      while (p0 != beg0)
+	if (*--p0 != *--p1)
 	  {
 	    /* Point at the first char of the matching suffix.  */
+	    ++p0, ++p1;
 	    beg0 = p0;
 	    break;
 	  }
